import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'package:http/http.dart' as http;

import 'package:shared_preferences/shared_preferences.dart';

typedef HttpProgressListener(int totalBytes, int receivedBytes, bool done);

class Http {
  static String baseURL;
  static String _token;
  static bool followRedirect = true;

  static Future<bool> loadAccessToken() async {
    SharedPreferences pref = await SharedPreferences.getInstance();
    if (pref.containsKey("token")) {
      _token = pref.getString("token");
      return true;
    }

    return false;
  }

  static Future<bool> setAccessToken(String token) async {
    _token = token;

    try {
      SharedPreferences pref = await SharedPreferences.getInstance();
      if (await pref.setString("token", token)) {
        return true;
      }
    } catch (e) {}

    return false;
  }

  static Future<bool> removeAccessToken() async {
    _token = null;
    try {
      SharedPreferences pref = await SharedPreferences.getInstance();
      if (pref.containsKey("token")) {
        await pref.remove("token");
        return true;
      }
    } catch (e) {}

    return false;
  }

  static _processHeader({
    HttpClientRequest request,
    Map<String, dynamic> headers,
    bool useAuthorization = true,
  }) {
    if (headers != null) {
      headers.forEach((key, value) {
        request.headers.add(key, value);
      });
    }

    if (useAuthorization && _token != null) {
      request.headers.set("Authorization", _token);
    }
  }

  static _processResponse(
      {HttpClientResponse response,
      HttpProgressListener progressListener}) async {
    final int totalBytes = response.contentLength;
    int receivedBytes = 0;

    String body = await response
        .transform(
          StreamTransformer.fromHandlers(
            handleData: (data, sink) {
              sink.add(data);

              if (progressListener != null) {
                receivedBytes += data.length;
                progressListener(totalBytes, receivedBytes, false);
              }
            },
            handleDone: (sink) {
              sink.close();
              if (progressListener != null) {
                progressListener(totalBytes, receivedBytes, true);
              }
            },
          ),
        )
        .map((v) => utf8.decoder.convert(v))
        .join();

    return body;
  }

  static _processResponseHttp(
      {http.StreamedResponse response,
      HttpProgressListener progressListener}) async {
    final int totalBytes = response.contentLength;
    int receivedBytes = 0;

    String body = await response.stream
        .transform(
          StreamTransformer.fromHandlers(
            handleData: (data, sink) {
              sink.add(data);

              if (progressListener != null) {
                receivedBytes += data.length;
                progressListener(totalBytes, receivedBytes, false);
              }
            },
            handleDone: (sink) {
              sink.close();
              if (progressListener != null) {
                progressListener(totalBytes, receivedBytes, true);
              }
            },
          ),
        )
        .map((v) => utf8.decoder.convert(v))
        .join();

    return body;
  }

  static raw({
    String url,
    Map<String, dynamic> headers,
    HttpProgressListener downloadProgressListener,
    String method,
  }) async {
    return await getData(
      baseURL: url,
      endpoint: "",
      headers: headers,
      downloadProgressListener: downloadProgressListener,
      method: method,
      useAuthorization: false,
      returnAsJson: false,
    );
  }

  static uploadFile({
    String baseURL,
    String endpoint,
    Map<String, dynamic> headers,
    bool useAuthorization = true,
    bool returnAsJson = true,
    File file,
    HttpProgressListener uploadProgressListener,
  }) async {
    String mBaseURL = baseURL ?? Http.baseURL;
    var uri = Uri.parse("$mBaseURL$endpoint");

    final request = http.MultipartRequest("POST", uri);

    final int fileSize = await file.length();
    int sentBytes = 0;
    final trf = file.openRead().transform<List<int>>(
          StreamTransformer.fromHandlers(
            handleData: (data, sink) {
              sink.add(data);

              if (uploadProgressListener != null) {
                sentBytes += data.length;
                uploadProgressListener(fileSize, sentBytes, false);
              }
            },
            handleDone: (sink) {
              sink.close();
              if (uploadProgressListener != null) {
                uploadProgressListener(fileSize, sentBytes, true);
              }
            },
          ),
        );

    request.files.add(
      http.MultipartFile(
        "file",
        trf,
        await file.length(),
      ),
    );

    if (headers != null) {
      headers.forEach((key, value) {
        request.headers[key] = value.toString();
      });
    }

    if (useAuthorization && _token != null) {
      request.headers["Authorization"] = _token;
    }

    final response = await request.send();
    if (response.statusCode == 200) {
      if (returnAsJson) {
        return json.decode(
          await _processResponseHttp(
            response: response,
            progressListener: null,
          ),
        );
      }

      return await _processResponseHttp(
        response: response,
        progressListener: null,
      );
    }

    return null;
  }

  static getData({
    String baseURL,
    String endpoint,
    Map<String, dynamic> headers,
    HttpProgressListener downloadProgressListener,
    dynamic data,
    String method,
    bool useAuthorization = true,
    bool returnAsJson = true,
  }) async {
    final client = HttpClient();
    try {
      client.userAgent = "AIFramework/";
      client.connectionTimeout = Duration(seconds: 30);

      HttpClientRequest request;
      String mBaseURL = baseURL ?? Http.baseURL;

      var uri = Uri.parse("$mBaseURL$endpoint");

      if (method != null && method.isNotEmpty) {
        switch (method.toUpperCase()) {
          case "GET":
            request = await client.getUrl(uri);
            break;

          case "POST":
            request = await client.postUrl(uri);
            break;

          case "PUT":
            request = await client.putUrl(uri);
            break;

          case "DELETE":
            request = await client.deleteUrl(uri);
            break;

          case "PATCH":
            request = await client.patchUrl(uri);
            break;

          case "HEAD":
            request = await client.headUrl(uri);
            break;

          default:
            request = await client.getUrl(uri);
            break;
        }
      } else {
        if (data == null) {
          request = await client.getUrl(uri);
        } else {
          request = await client.postUrl(uri);
        }
      }

      request.followRedirects = followRedirect;

      _processHeader(
        request: request,
        headers: headers,
        useAuthorization: useAuthorization,
      );

      if (data != null) {
        request.headers.add("Content-Type", "application/json");

        String payload = json.encode(data);
        Uint8List bytes = utf8.encode(payload);
        request.headers.add("Content-Length", bytes.lengthInBytes.toString());

        request.add(bytes);
        await request.flush();
      }

      final response = await request.close();
      if (response.statusCode == 200) {
        if (returnAsJson) {
          return json.decode(
            await _processResponse(
              response: response,
              progressListener: downloadProgressListener,
            ),
          );
        }

        return await _processResponse(
          response: response,
          progressListener: downloadProgressListener,
        );
      }
    } finally {
      client.close(
        force: true,
      );
    }

    return null;
  }
}
